from nose.tools import assert_raises, with_setup
from os import remove
from os.path import exists

from pkglts.hash_management import compute_hash, modified_file_hash, pth_as_key


ex_file = 'toto.txt'


def teardown():
    if exists(ex_file):
        remove(ex_file)


def test_pth_as_key_produce_unique_path():
    pth1 = pth_as_key("./toto/titi/")
    pth2 = pth_as_key("toto/titi")
    pth3 = pth_as_key("./toto/../toto/titi")

    assert pth1 == pth2
    assert pth1 == pth3


def test_compute_hash():
    assert compute_hash("toto") == compute_hash("toto")
    assert compute_hash("toto") != compute_hash("titi")


@with_setup(teardown=teardown)
def test_modified_file_raise_error_if_file_not_in_hashmap():
    with open(ex_file, 'w') as f:
        f.write("lorem ipsum\n" * 10)

    assert_raises(IOError, lambda: modified_file_hash(ex_file, {}))


@with_setup(teardown=teardown)
def test_modified_file_detect_modifications_only_in_preserved_sections():
    with open(ex_file, 'w') as f:
        f.write("lorem ipsum\n" * 10)

    assert not modified_file_hash(ex_file, {ex_file: []})


@with_setup(teardown=teardown)
def test_modified_file_if_not_same_preserved_sections():
    txt = "lorem ipsum\n" * 10

    with open(ex_file, 'w') as f:
        f.write("{# pkglts, toto\n")
        f.write(txt)
        f.write("#}\n")

    assert modified_file_hash(ex_file, {ex_file: dict(titi="azerty")})


@with_setup(teardown=teardown)
def test_modified_file_detect_modifications_in_preserved_sections():
    txt = "lorem ipsum\n" * 10
    hv = compute_hash(txt)

    with open(ex_file, 'w') as f:
        f.write("{# pkglts, toto\n")
        f.write(txt)
        f.write("\n#}\n")

    assert not modified_file_hash(ex_file, {ex_file: dict(toto=hv)})

    with open(ex_file, 'w') as f:
        f.write("{# pkglts, toto\n")
        f.write(txt * 2)
        f.write("\n#}\n")

    assert modified_file_hash(ex_file, {ex_file: dict(toto=hv)})



parameters = [
    ("intended_versions", ["27"]),
    ("require", [])
]


def check(env):
    """Check the validity of parameters in working environment.

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    intended_versions = env.globals['pysetup'].intended_versions

    if len(intended_versions) == 0:
        invalids.append("intended_versions")

    require = env.globals['pysetup'].require

    valid_methods = ("none", "pip", "conda", "git")
    if any(imeth not in valid_methods for imeth, name in require):
        invalids.append("require")

    return invalids



from nose.tools import assert_raises

from pkglts.config_management import create_env
from pkglts.local import (pkg_full_name, src_dir)


print(__file__)


def test_pkg_full_name():
    cfg = {}
    assert_raises(KeyError, lambda: pkg_full_name(create_env(cfg)))

    cfg['base'] = dict(pkgname='toto', namespace=None)
    name1 = pkg_full_name(create_env(cfg))
    cfg['base']['namespace'] = 'oa'
    name2 = pkg_full_name(create_env(cfg))

    assert name1 != name2


def test_src_dir():
    cfg = {}
    assert_raises(KeyError, lambda: src_dir(create_env(cfg)))

    cfg['base'] = dict(pkgname='toto', namespace=None)
    dir1 = src_dir(create_env(cfg))
    cfg['base']['namespace'] = 'oa'
    dir2 = src_dir(create_env(cfg))

    assert dir1 != dir2



from pkglts.local import pkg_full_name
from pkglts.option.doc import fmt_badge


def auto_classifiers(env):
    items = set(env.globals['pypi'].classifiers)

    # add license item
    # TODO

    # add intended versions items
    intended_versions = env.globals['pysetup'].intended_versions
    if len(intended_versions) > 0:
        items.add("Programming Language :: Python")

        ver_cla_tpl = "Programming Language :: Python :: %s.%s"
        major_versions = set()
        for ver in intended_versions:
            items.add(ver_cla_tpl % (ver[0], ver[1]))
            major_versions.add(ver[0])

        ver_cla_tpl = "Programming Language :: Python :: %s"
        for ver in major_versions:
            items.add(ver_cla_tpl % ver)

        if len(major_versions) == 1:
            ver, = major_versions
            items.add("Programming Language :: Python :: %s :: Only" % ver)

    return sorted(items)


def environment_extensions(env):
    """Add more functionality to an environment.

    Args:
        env (jinja2.Environment):

    Returns:
        dict of str: any
    """
    url = "badge.fury.io/py/%s" % pkg_full_name(env)
    img = url + ".svg"
    badge = fmt_badge(img, url, "PyPI version")

    return {"badge": badge,
            "auto_classifiers": auto_classifiers(env)}



from os.path import exists

from .handlers import get_tpl_path


parameters = [
    ("name", "cecill-c"),
    ("year", 2015),
    ("organization", "organization"),
    ("project", "{{ base.pkgname }}")
]


def check(env):
    """Check the validity of parameters in working environment.

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    name = env.globals['license'].name
    # year = pkg_cfg['license']['year']
    # organization = pkg_cfg['license']['organization']
    # project = pkg_cfg['license']['project']

    if len(name) == 0:
        invalids.append('name')
    elif not exists(get_tpl_path(name)):
        invalids.append('name')

    return invalids



from nose.tools import with_setup

from pkglts.install_env.pip_front_end import install, installed_packages

from .venv_tools import clear_venv, create_venv

__test__ = False

venv = "tyti_pip"
mem = {}


def setup_func():
    create_venv(venv, mem)


def teardown_func():
    clear_venv(venv, mem)


def test_installed_packages():
    pkgs = set(installed_packages())
    assert "wheel" in pkgs


@with_setup(setup_func, teardown_func)
def test_install():
    # create virtualenv
    assert "my-first-p" not in installed_packages()

    # install local wheel
    install("test/test_install_env/my_first_p-0.0.4-py2.py3-none-any.whl")

    # test new package still not in virtualenv
    assert "my-first-p" in installed_packages()



from pkglts.config_management import create_env
from pkglts.option.readthedocs.config import check, parameters


def test_parameters():
    assert len(parameters) == 1


def test_config_check_project_exists():
    env = create_env(dict(readthedocs={'project': ""}))
    assert 'project' in check(env)



from funcsigs import signature
from inspect import getsource
from uuid import uuid1

from .parse_doc import parse_docstring


def create_node_def(func):
    """Create a node definition schema for the given function

    Args:
        func: (function) actual function object

    Returns:
        (dict): schema as specified in node/schema
    """
    node_def = dict(id=uuid1().hex,
                    name=func.__name__,
                    description=func.__doc__,
                    author="unknown",
                    version=0,
                    function="%s:%s" % (func.__module__, func.__name__),
                    inputs=[],
                    outputs=[])

    s = signature(func)
    pycode = getsource(func)
    parsed = parse_docstring(func.__doc__)
    for pname, p in s.parameters.items():
        if pname in parsed['args']:
            arg_type, descr = parsed['args'][pname]
        else:
            arg_type = "any"
            descr = ""

        if p.annotation == p.empty:
            interface = arg_type
        else:
            interface = str(p.annotation)

        description = str(descr)

        port = dict(name=pname, interface=interface, description=description)
        if p.default != p.empty:
            port["default"] = str(p.default)
        node_def['inputs'].append(port)

    if "return" in pycode:

        if parsed['returns'] is not None:
            ret_name, ret_type, ret_descr = parsed['returns']
            if ret_name is None:
                ret_name = "ret"

            if s.return_annotation == s.empty:
                if ret_type is None:
                    interface = "IAny"
                else:
                    interface = ret_type
            else:
                interface = str(s.return_annotation)

            if ret_descr is None:
                description = ""
            else:
                description = str(ret_descr)

            port = dict(name=ret_name,
                        interface=interface,
                        description=description)

            node_def['outputs'].append(port)

    return node_def



from os.path import abspath, basename


parameters = [
    ("pkgname", basename(abspath("."))),
    ("namespace", None),
    ("url", None),
    ("authors", [("moi", "moi@email.com")])
]


def is_valid_identifier(name):
    """ Check that name is a valid python identifier
    sort of back port of "".isidentifier()
    """
    try:
        compile("%s=1" % name, "test", 'single')
        return True
    except SyntaxError:
        return False


def check(env):
    """Check the validity of parameters in working environment.

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    pkgname = env.globals['base'].pkgname
    namespace = env.globals['base'].namespace

    if "." in pkgname:
        invalids.append('pkgname')
    elif not is_valid_identifier(pkgname):
        invalids.append('pkgname')

    if namespace is not None:
        if "." in namespace:
            invalids.append('namespace')
        elif not is_valid_identifier(namespace):
            invalids.append('namespace')

    return invalids


def after(pkg_cfg):
    del pkg_cfg  # unused
    print("base: after main config")



from pkglts.config_management import create_env
from pkglts.manage import default_cfg
from pkglts.manage_tools import check_option_parameters


def test_check_handle_option_config_not_existing():
    env = create_env(default_cfg)
    assert len(check_option_parameters("toto", env)) == 0


def test_check_handle_option_defines_no_check():
    env = create_env(default_cfg)
    assert len(check_option_parameters("test", env)) == 0



from pkglts.option.doc import fmt_badge


def environment_extensions(env):
    """Add more functionality to an environment.

    Args:
        env (jinja2.Environment):

    Returns:
        dict of str: any
    """
    project = env.globals['readthedocs'].project
    project = project.replace(".", "")
    url = "%s.readthedocs.io/en/latest/?badge=latest" % project
    img = "readthedocs.org/projects/%s/badge/?version=latest" % project
    badge = fmt_badge(img, url, "Documentation status")

    return {"badge": badge}



from nose.tools import with_setup
from os import remove
from os.path import exists

from pkglts.file_management import write_file


ex_file = 'toto.txt'


def teardown():
    if exists(ex_file):
        remove(ex_file)


@with_setup(teardown=teardown)
def test_write_file():
    write_file(ex_file, "lorem ipsum")
    assert exists(ex_file)




from copy import deepcopy

from pkglts.config_management import pkg_env


def test_pkg_url_empty_default():
    env = pkg_env(dict(pysetup={"intended_versions": ["27"],
                                'require': []}))
    assert env.globals['pysetup'].pkg_url == ""


def test_pkg_url_look_multiple_places():
    cfg = dict(base={'pkgname': 'toto', 'namespace': 'oa', 'url': None},
               github={'url': None},
               pypi={'classifiers': [], 'url': None},
               readthedocs={'project': 'project'},
               pysetup={'intended_versions': ["27"],
                        'require': []})

    for name in ("base", "github", "pypi", "readthedocs"):
        loc_cfg = deepcopy(cfg)
        loc_cfg[name]['url'] = name
        env = pkg_env(loc_cfg)
        assert env.globals['pysetup'].pkg_url == name


def test_requirements():
    cfg = dict(test={},
               pysetup={'intended_versions': ["27"],
                        'require': []})
    env = pkg_env(cfg)
    assert len(env.globals['pysetup'].requirements('install')) == 0
    assert len(env.globals['pysetup'].requirements('dvlpt')) == 2




from pkglts.config_management import pkg_env


def test_badge():
    env = pkg_env(dict(coveralls={},
                          github={'owner': "moi", 'project': "project"}))
    assert ".. image:" in env.globals['coveralls'].badge



from pkglts.config_management import create_env
from pkglts.option.version.config import check, parameters


def test_parameters():
    assert len(parameters) == 3


def test_config_check_version_numbers_are_valid():
    env = create_env(dict(version={'major': "", 'minor': "", 'post': ""}))
    assert 'major' in check(env)
    assert 'minor' in check(env)
    assert 'post' in check(env)
    env = create_env(dict(version={'major': "a", 'minor': "a", 'post': "a"}))
    assert 'major' in check(env)
    assert 'minor' in check(env)
    assert 'post' in check(env)
    env = create_env(dict(version={'major': "1", 'minor': "1", 'post': "1"}))
    assert 'major' in check(env)
    assert 'minor' in check(env)
    assert 'post' in check(env)
    env = create_env(dict(version={'major': 1, 'minor': 0, 'post': "2.dev"}))
    assert 'post' in check(env)



parameters = [
    ("major", 0),
    ("minor", 1),
    ("post", 0)
]


def check(env):
    """Check the validity of parameters in working environment.

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    major = env.globals['version'].major
    minor = env.globals['version'].minor
    post = env.globals['version'].post

    if not isinstance(major, int):
        invalids.append("major")
    if not isinstance(minor, int):
        invalids.append("minor")
    if not isinstance(post, int):
        invalids.append("post")

    return invalids



"""Regroup set of functions that make use of local environment
inside a package. Just a way to normalize pre defined paths.
"""

from os.path import exists

from .file_management import write_file
from .templating import closing_marker, opening_marker


def pkg_full_name(env):
    """Compute name of src dir according to pkgname
    and namespace in environment

    Args:
        env (jinja2.Environment): current working environment

    Returns:
        (str)
    """
    namespace = env.globals['base'].namespace
    if namespace is None:
        return env.globals['base'].pkgname
    else:
        return namespace + "." + env.globals['base'].pkgname


def src_dir(env):
    """Compute name of src dir according to pkgname
    and namespace in environment

    Args:
        env (jinja2.Environment): current working environment

    Returns:
        (str)
    """
    rep = "src"
    namespace = env.globals['base'].namespace
    if namespace is not None:
        rep = rep + "/" + namespace

    pkgname = env.globals['base'].pkgname
    rep = rep + "/" + pkgname

    return rep


namespace_txt = """
# %spkglts,
__import__('pkg_resources').declare_namespace(__name__)
# %s
""" % (opening_marker, closing_marker)


def init_namespace_dir(pth):
    """Populate a directory with specific __init__.py
    for namespace packages.

    Args:
        pth (str): path in which to create the files

    Returns:
        None
    """
    init_pth = pth + "/__init__.py"
    if not exists(init_pth):
        write_file(init_pth, namespace_txt)



import logging
from os.path import dirname
from os.path import join as pj


logger = logging.getLogger(__name__)

tpl_dir = pj(dirname(__file__), "templates")


def get_tpl_path(name):
    """Return a path for license template file.

    Warnings: Do not test if path exists

    Args:
        name (str): name of license to fetch

    Returns:
        (str)
    """
    tpl_pth = pj(tpl_dir, "%s.txt" % name.lower())
    return tpl_pth


def full_text(env):
    """Generate a license
    """
    name = env.globals['license'].name

    # open template
    try:
        with open(get_tpl_path(name), 'r') as f:
            cnt = f.read()

        return env.from_string(cnt).render()
    except IOError as e:
        logger.error("unable to find template for given license")
        raise e


def environment_extensions(env):
    """Add more functionality to an environment.

    Args:
        env (jinja2.Environment):

    Returns:
        dict of str: any
    """
    return {"full_text": full_text(env)}



from pkglts.config_management import pkg_env


def test_badge():
    env = pkg_env(dict(base={'pkgname': "pkg", 'namespace': None,
                             "url": "http://toto"},
                       pysetup={"intended_versions": ["27", "28"],
                                'require': []},
                       pypi={'classifiers': []}))
    assert ".. image:" in env.globals['pypi'].badge


def test_auto_classifiers():
    env = pkg_env(dict(base={'pkgname': "pkg", 'namespace': None,
                             "url": "http://toto"},
                       pysetup={"intended_versions": ["27", "28"],
                                'require': []},
                       pypi={'classifiers': []}))
    section = env.globals['pypi']
    assert "Programming Language :: Python" in section.auto_classifiers
    assert "Programming Language :: Python :: 2.7" in section.auto_classifiers



"""Pip front end to install package in current environment.
"""
from json import loads
from subprocess import Popen, PIPE


def installed_packages():
    """Iterate on all packages installed in the current python environment.

    return:
        (iter of str)
    """
    p = Popen(["conda", "list", "--json"], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    output, err = p.communicate("")
    rc = p.returncode
    if rc != 0:
        raise UserWarning("unable to execute 'conda list'")

    for pkg in loads(output):
        yield pkg.split("-")[0]


def install(name):
    """Install a package in the current python environment.

    arg:
     - name (str): name of the package

    return:
     - (bool): whether installation was successful or not
    """
    p = Popen(["conda", "install", name], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    p.communicate("")
    rc = p.returncode

    if rc == 0:
        return True
    else:  # try to use pip if conda install fail
        from .pip_front_end import install as pip_install
        return pip_install(name)



import json
import mock
from nose.tools import assert_raises, with_setup
from os import remove
from os.path import exists
from os.path import join as pj
from nose.tools import assert_raises

from pkglts.config_management import (create_env, current_pkg_cfg_version,
                                      default_cfg, get_pkg_config,
                                      installed_options, pkg_env,
                                      write_pkg_config)

from .small_tools import ensure_created, rmdir


tmp_dir = 'toto_mg_cfg'


def setup():
    ensure_created(tmp_dir)
    ensure_created(pj(tmp_dir, ".pkglts"))


def teardown():
    rmdir(tmp_dir)


def test_create_env():
    env = create_env(default_cfg)
    assert len(tuple(installed_options(env))) == 0

    cfg = dict(default_cfg)
    cfg['base'] = dict(a=1, b=2)
    env = create_env(cfg)
    assert len(tuple(installed_options(env))) == 1
    assert env.globals['base'].a == 1
    assert env.globals['base'].b == 2


def test_create_env_render_templates():
    cfg = dict(default_cfg)
    cfg['base'] = dict(a="a", b="b")
    cfg['tpl'] = dict(tpl1="{{ base.a }}",
                      tpl2="{{ base.b }} and {{ tpl.tpl1 }}")

    env = create_env(cfg)
    assert env.globals['tpl'].tpl1 == "a"
    assert env.globals['tpl'].tpl2 == "b and a"


def test_create_env_raise_error_if_unable_to_fully_render_templates():
    cfg = dict(default_cfg)

    cfg['base'] = dict(a="{{ base.b }}", b="{{ base.a }}")
    assert_raises(UserWarning, lambda: create_env(cfg))


def test_pkg_env_raise_error_if_option_not_defined():
    cfg = dict(default_cfg)
    cfg['babou'] = dict(a="a", b="b")

    assert_raises(KeyError, lambda: pkg_env(cfg))


def test_pkg_env_loads_specific_handlers_from_options():
    cfg = dict(default_cfg)
    cfg['base'] = dict(pkgname="toto", namespace="nm",
                       url=None, authors=[("moi", "moi@aussi")])

    env = pkg_env(cfg)
    assert hasattr(env.globals['base'], "pkg_full_name")


def test_get_pkg_config_read_cfg():
    cfg = dict(default_cfg)
    cfg['base'] = dict(pkgname="toto", namespace="nm",
                       url=None, authors=[("moi", "moi@aussi")])
    with open(pj(tmp_dir, ".pkglts/pkg_cfg.json"), 'w') as f:
        json.dump(cfg, f)

    env = get_pkg_config(tmp_dir)
    assert 'base' in env.globals


def test_get_pkg_config_handle_versions():
    cfg = dict(default_cfg)
    cfg["_pkglts"]["version"] = 0
    with open(pj(tmp_dir, ".pkglts/pkg_cfg.json"), 'w') as f:
        json.dump(cfg, f)

    env = get_pkg_config(tmp_dir)
    assert env.globals["_pkglts"].version == current_pkg_cfg_version


def test_pkg_cfg_read_write_maintains_templates():
    cfg = dict(default_cfg)
    cfg['base'] = dict(pkgname="toto", namespace="nm",
                       url=None, authors=[("moi", "moi@aussi")])
    cfg['license'] = dict(name="CeCILL-C", organization="org",
                          project="{{ base.pkgname }}", year="2015")

    with open(pj(tmp_dir, ".pkglts/pkg_cfg.json"), 'w') as f:
        json.dump(cfg, f)

    env = get_pkg_config(tmp_dir)
    assert env.globals['license'].project == "toto"

    write_pkg_config(env, tmp_dir)
    with open(pj(tmp_dir, ".pkglts/pkg_cfg.json"), 'r') as f:
        cfg = json.load(f)

    cfg["base"]["pkgname"] = "tutu"

    with open(pj(tmp_dir, ".pkglts/pkg_cfg.json"), 'w') as f:
        json.dump(cfg, f)

    env = get_pkg_config(tmp_dir)
    assert env.globals['license'].project == "tutu"



""" Specific helper function for manage script
"""

from importlib import import_module
import logging
from os import listdir, mkdir
from os.path import basename, exists, isdir, splitext

from .config_management import ConfigSection, installed_options
from .hash_management import compute_hash, pth_as_key
from .install_env.load_front_end import get_install_front_end
from .local import init_namespace_dir
from .option_tools import get_user_permission
from .templating import render


logger = logging.getLogger(__name__)

tpl_src_name = "{" + "{ base.pkgname }" + "}"

non_bin_ext = ("", ".bat", ".cfg", ".in", ".ini", ".no", ".py", ".rst", ".sh",
               ".txt", ".yml", ".yaml")


def ensure_installed_packages(requirements, msg, env):
    """Ensure all packages in requirements are installed.

    If not, ask user permission to install them.

    Args:
        requirements (list of str): list of package names to install
                                   if needed
        msg (str): error message to print
        env (jinja2.Environment): current working environment

    Returns:
        (bool): whether all required packages are installed or not
    """
    ife_name = env.globals["_pkglts"].install_front_end
    req = [name for repo, name in requirements
           if repo is None or repo == ife_name]
    ife = get_install_front_end(ife_name)
    to_install = set(req) - set(ife.installed_packages())
    if len(to_install) > 0:
        print(msg)
        logger.warning("missing packages: " + ", ".join(to_install))
        for name in to_install:
            ife.install(name)
            logger.info("install %s", name)

    return True


def check_option_parameters(name, env):
    """Check that the parameters associated to an option are valid.

    Try to import Check function in option dir.

    Args:
        name (str): option name
        env (jinja2.Environment): current working environment
    """
    try:
        opt_cfg = import_module("pkglts.option.%s.config" % name)
        try:
            return opt_cfg.check(env)
        except AttributeError:
            return []
    except ImportError:
        return []


def update_opt(name, env):
    """Update an option of this package.

    Notes: If the option does not exists yet, add it first.
           See the list of available option online

    Args:
        name (str): name of option to add
        env (jinja2.Environment): current working environment
    """
    logger.info("update option %s", name)

    # test existence of option
    try:
        opt_require = import_module("pkglts.option.%s.require" % name)
        opt_cfg = import_module("pkglts.option.%s.config" % name)
    except ImportError:
        raise KeyError("option '%s' does not exists" % name)

    # find other option requirements in repository
    for option_name in opt_require.option:
        if option_name not in installed_options(env):
            print("need to install option '%s' first" % option_name)
            if (env.globals["_pkglts"].auto_install or
                    get_user_permission("install")):
                env = update_opt(option_name, env)
            else:
                return env

    # find extra package requirements for setup
    msg = "this option requires some packages to setup"
    if not ensure_installed_packages(opt_require.setup, msg, env):
        print("option installation stopped")
        return env

    # find parameters required by option config
    try:
        params = opt_cfg.parameters
    except AttributeError:
        params = []

    option_cfg = ConfigSection()
    prev_cfg = env.globals.get(name, {})
    for key, default in params:
        option_cfg.add_param(key, getattr(prev_cfg, key, default))

    # write new pkg_info file
    env.globals[name] = option_cfg

    try:  # TODO: proper developer doc to expose this feature
        opt_cfg.after(env)
    except AttributeError:
        pass

    # find extra package requirements for dvlpt
    msg = "this option requires additional packages for developers"
    ensure_installed_packages(opt_require.dvlpt, msg, env)

    return env


def regenerate_dir(src_dir, tgt_dir, env, overwrite_file):
    """Walk all files in src_dir and create/update them on tgt_dir

    Args:
        src_dir (str): path to reference files
        tgt_dir (str): path to target where files will be written
        env (jinja2.Environment): current working environment
        overwrite_file (dict of str, bool): whether or not to overwrite some
                             files

    Returns:
        (dict of str, map): hash key of preserved sections
    """
    hm = {}

    for src_name in listdir(src_dir):
        src_pth = src_dir + "/" + src_name
        tgt_name = env.from_string(src_name).render()
        if tgt_name.endswith(".tpl"):
            tgt_name = tgt_name[:-4]

        tgt_pth = tgt_dir + "/" + tgt_name
        # handle namespace
        if (isdir(src_pth) and basename(src_dir) == 'src' and
                    src_name == tpl_src_name):
            namespace = env.globals['base'].namespace
            if namespace is not None:
                ns_pth = tgt_dir + "/" + namespace
                if not exists(ns_pth):
                    mkdir(ns_pth)

                init_namespace_dir(ns_pth)
                tgt_pth = ns_pth + "/" + tgt_name

        if isdir(src_pth):
            if tgt_name not in ("", "_") and not exists(tgt_pth):
                mkdir(tgt_pth)

            sub_hm = regenerate_dir(src_pth, tgt_pth, env, overwrite_file)
            hm.update(sub_hm)
        else:
            kp = pth_as_key(tgt_pth)
            if overwrite_file.get(kp, True):
                fname, ext = splitext(tgt_name)
                if ext in non_bin_ext:
                    blocks = render(env, src_pth, tgt_pth)
                    hm[kp] = dict((bid, compute_hash(cnt)) for bid, cnt in blocks)
                else:  # binary file
                    if exists(tgt_pth):
                        print("overwrite? %s" % tgt_pth)
                    else:
                        with open(src_pth, 'rb') as fr:
                            content = fr.read()
                        with open(tgt_pth, 'wb') as fw:
                            fw.write(content)

    return hm



from pkglts.config_management import create_env
from pkglts.option.pysetup.config import check, parameters


def test_parameters():
    assert len(parameters) == 2


def test_config_check_intended_version_exists():
    env = create_env(dict(pysetup={'intended_versions': [], 'require': []}))
    assert 'intended_versions' in check(env)
    assert 'require' not in check(env)

    env = create_env(dict(pysetup={'intended_versions': ["27"],
                                   'require': [('walou', 'numpy')]}))
    assert 'require' in check(env)



from copy import deepcopy
from nose.tools import with_setup
from os import listdir
from os.path import exists
from os.path import join as pj

from pkglts.config_management import create_env, default_cfg, pkg_env
from pkglts.manage import install_example_files

from .small_tools import ensure_created, rmdir


tmp_dir = "tmp_cfgex"


def setup_func():
    ensure_created(tmp_dir)


def teardown_func():
    rmdir(tmp_dir)


@with_setup(setup_func, teardown_func)
def test_install_example_returns_false_if_option_not_already_installed():
    env = create_env(default_cfg)
    ans = install_example_files('option', env, tmp_dir)
    assert not ans


@with_setup(setup_func, teardown_func)
def test_install_example_ok_if_option_do_not_provide_examples():
    cfg = deepcopy(default_cfg)
    cfg['base'] = dict()
    env = create_env(cfg)
    ans = install_example_files('base', env, tmp_dir)
    assert not ans


@with_setup(setup_func, teardown_func)
def test_install_example_copy_files():
    cfg = deepcopy(default_cfg)
    cfg['base'] = dict(pkgname='toto', namespace=None)
    cfg['test'] = dict()
    env = pkg_env(cfg)

    assert len(listdir(tmp_dir)) == 0
    install_example_files('test', env, tmp_dir)
    assert len(listdir(tmp_dir)) > 0
    assert exists(pj(tmp_dir, "src", "toto", "example.py"))
    assert exists(pj(tmp_dir, "test", "test_example.py"))


@with_setup(setup_func, teardown_func)
def test_install_example_copy_binary_files():
    cfg = deepcopy(default_cfg)
    cfg['base'] = dict(pkgname='toto', namespace=None)
    cfg['data'] = dict()
    env = pkg_env(cfg)

    assert len(listdir(tmp_dir)) == 0
    install_example_files('data', env, tmp_dir)
    assert len(listdir(tmp_dir)) > 0
    assert exists(pj(tmp_dir, "src", "toto_data", "ext_data.png"))


@with_setup(setup_func, teardown_func)
def test_install_example_do_not_complain_if_file_already_exists():
    cfg = deepcopy(default_cfg)
    cfg['base'] = dict(pkgname='toto', namespace=None)
    cfg['test'] = dict()
    env = pkg_env(cfg)

    assert install_example_files('test', env, tmp_dir)
    assert install_example_files('test', env, tmp_dir)


@with_setup(setup_func, teardown_func)
def test_install_example_handles_namespace():
    cfg = deepcopy(default_cfg)
    cfg['base'] = dict(pkgname='toto', namespace='oa')
    cfg['test'] = dict()
    env = pkg_env(cfg)

    install_example_files('test', env, tmp_dir)
    assert exists(pj(tmp_dir, "test", "test_example.py"))
    assert not exists(pj(tmp_dir, "src", "toto", "example.py"))
    assert exists(pj(tmp_dir, "src", "oa", "toto", "example.py"))



import re
from sphinx.util.docstrings import prepare_docstring


def parse_docstring(txt):
    """Extract components of a google formatted docstring

    Args:
        txt: (str) actual docstring to parse

    Returns:
        (dict of (str, val)) mapping of docstring sections
    """
    doc = [line.rstrip() for line in prepare_docstring(txt) if len(line.strip()) > 0]

    parsed = dict(caption="", doc=[], args={}, returns=None)

    arg_pattern = re.compile(r"""
    ^(?P<name>\w+)       # param name
    (:[ ]*               # rest is optional (only param name)
    (                    # param type is optional
    \((?P<type>\w+)\)    # param type
    )?                   # optional param type
    [ ]*(?P<descr>.+$)?  # param description consume the rest of the line
    )?                   # case of only param name
    """, re.VERBOSE)

    ret_pattern = re.compile(r"""
    ^((?P<name>\w+):[ ]*)?      # return name is optional
    \((?P<type>\w+)\)    # return type
    [ ]*(?P<descr>.+$)?  # return description consume the rest of the line
    """, re.VERBOSE)

    state = ["top"]
    for line in doc:
        if line.startswith("Args:"):
            if state[-1] != "top":
                state.pop(-1)

            state.append("args")
        elif line.startswith("Returns:"):
            if state[-1] != "top":
                state.pop(-1)

            state.append("returns")
        else:
            if state[-1] == "top":
                if parsed['caption'] == "":
                    parsed['caption'] = line
                else:
                    parsed['doc'].append(line)
            elif state[-1] == "args":
                if line[0] != " ":
                    state.pop(-1)
                    parsed['doc'].append(line)
                else:
                    match = re.match(arg_pattern, line.lstrip())
                    parsed['args'][match.group('name')] = (match.group('type'),
                                                           match.group('descr'))
            elif state[-1] == "returns":
                if line[0] != " ":
                    state.pop(-1)
                    parsed['doc'].append(line)
                else:
                    match = re.match(ret_pattern, line.lstrip())
                    parsed['returns'] = (match.group('name'),
                                        match.group('type'),
                                        match.group('descr'))
            else:
                raise UserWarning("how did we even reach this point?")

    return parsed



from importlib import import_module
from inspect import isfunction
import json
from os.path import exists, splitext
from pkgutil import walk_packages

from pkglts.local import pkg_full_name

from .create_node_def import create_node_def


def find_plugin_categories(txt):
    if txt is None:
        return []

    if '__plugin__' not in txt:
        return []

    for line in txt.splitlines():
        if '__plugin__' in line:
            cat_txt = line.strip().split(":")[1].strip()
            return [c.strip() for c in cat_txt.split(",")]


def parse_plugins(pkg):
    """Parse recursively all plugins in a given package

    Notes: write plugin def files

    Args:
        pkg: (Package) a python package object

    Returns:
        None
    """
    for imp, modname, ispkg in walk_packages(pkg.__path__):
        loader = imp.find_module(modname)
        mod = import_module("%s.%s" % (pkg.__name__, loader.fullname))
        if ispkg:
            parse_plugins(mod)
        else:
            root_pth = splitext(mod.__file__)[0]
            plugins = {}
            if hasattr(mod, '__all__'):
                item_names = mod.__all__
            else:
                item_names = dir(mod)
            for item_name in item_names:
                item = getattr(mod, item_name)
                if isfunction(item):
                    for cat in find_plugin_categories(item.__doc__):
                        if cat == "node":
                            pth = "%s_%s.wkf" % (root_pth, item_name)
                            plugins[item_name] = (pth, create_node_def(item))

            if len(plugins) > 0:
                for item_name, (pth, idef) in plugins.items():
                    if exists(pth):
                        with open(pth, 'r') as f:
                            old_idef = json.load(f)

                        idef['id'] = old_idef['id']

                    with open(pth, 'w') as f:
                        json.dump(idef, f, indent=2)


def main(env, target=".", overwrite=False):
    """Main function called to walk the package

    Args:
        env: (dict of (str, dict)) package configuration parameters
        target (str): place to write plugin def into
        overwrite (bool): whether or not to overwrite previous definition
                          files. Default to False.
    """
    del target
    del overwrite
    pkg = import_module(pkg_full_name(env))
    parse_plugins(pkg)



from pkglts.local import pkg_full_name, src_dir


def environment_extensions(env):
    """Add more functionality to an environment.

    Args:
        env (jinja2.Environment):

    Returns:
        dict of str: any
    """
    return {"pkg_full_name": pkg_full_name(env),
            "src_pth": src_dir(env)}



from nose.tools import assert_raises

from pkglts.config_management import pkg_env


def test_generate():
    cfg = dict(name='mit',
               year="2015",
               organization="org",
               project="project")
    env = pkg_env(dict(license=cfg))
    assert len(env.globals['license'].full_text) > 0


def test_generate_raise_error_if_license_do_not_exists():
    cfg = dict(name='tugudu',
               year="2015",
               organization="org",
               project="project")
    assert_raises(IOError, lambda: pkg_env(dict(license=cfg)))



from pkglts.config_management import pkg_env


def test_src_pth():
    env = pkg_env({'base': {'pkgname': 'toto', 'namespace': None}})
    assert env.globals['base'].src_pth == "src/toto"
    env = pkg_env({'base': {'pkgname': 'toto', 'namespace': 'oa'}})
    assert env.globals['base'].src_pth == "src/oa/toto"


def test_pkg_full_name():
    env = pkg_env({'base': {'pkgname': 'toto', 'namespace': None}})
    assert env.globals['base'].pkg_full_name == "toto"
    env = pkg_env({'base': {'pkgname': 'toto', 'namespace': 'oa'}})
    assert env.globals['base'].pkg_full_name == "oa.toto"



parameters = [
    ("classifiers", [
            "Development Status :: 2 - Pre-Alpha",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: BSD License",
            "Natural Language :: English"
        ])
]


def check(env):
    """Check the validity of parameters in working environment.

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    classifiers = env.globals['pypi'].classifiers

    if len(classifiers) == 0:
        invalids.append("classifiers")

    return invalids



parameters = [
    ("project", "{{ github.project }}")
]


def check(env):
    """Check the validity of parameters in working environment.

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    project = env.globals['readthedocs'].project

    if len(project) == 0:
        invalids.append("project")

    return invalids



from os.path import basename

from pkglts.data_access import get, get_data_dir, ls


def test_data_access_data_dir():
    assert basename(get_data_dir()) == "pkglts_data"


def test_data_access_ls():
    assert (set(ls('test/test1')) ==
            {('subtest', True), ('titi.txt', False)})


def test_data_access_get():
    assert get('test/toto.txt').rstrip() == "lorem ipsum"



from importlib import import_module
from jinja2 import Environment, StrictUndefined, UndefinedError
import json
import logging
from os.path import join as pj

from .config import pkglts_dir, pkg_cfg_file

try:
    string_type = basestring
except NameError:
    string_type = str

current_pkg_cfg_version = 3

logger = logging.getLogger(__name__)

default_cfg = dict(_pkglts=dict(use_prompts=False,
                                auto_install=True,
                                install_front_end='stdout',
                                version=current_pkg_cfg_version))


class FormattedString(str):
    """Small class to hold both formatted string and its template
    """
    pass


class ConfigSection(object):
    """Small class to gather config keys
    """
    def __init__(self):
        self._params = []

    def add_param(self, key, val):
        """Add a new attribute to this section

        Args:
            key (str): python valid identifier
            val (any): actual value of the parameter

        Returns:
            (None)
        """
        self._params.append(key)
        setattr(self, key, val)

    def items(self):
        """Iterates on couple key, values stored as attributes

        Returns:
            (iter of {str, any})
        """
        for key in self._params:
            yield key, getattr(self, key)


def create_env(pkg_cfg):
    """Create a jinja2.Environment from a package configuration

    Notes: do not load option handlers

    Args:
        pkg_cfg (dict of str: any): package configuration

    Returns:
        (jinja2.Environment)
    """
    env = Environment(undefined=StrictUndefined)
    env.keep_trailing_newline = True

    to_eval = []
    for opt_name, cfg in pkg_cfg.items():
        env.globals[opt_name] = ConfigSection()
        for key, param in cfg.items():
            if isinstance(param, string_type):
                to_eval.append((opt_name, key, param))
            else:
                env.globals[opt_name].add_param(key, param)

    nb_iter_max = len(to_eval) ** 2
    cur_iter = 0
    while len(to_eval) > 0 and cur_iter < nb_iter_max:
        cur_iter += 1
        opt_name, key, param = to_eval.pop(0)
        try:
            txt = env.from_string(param).render()
            if txt != param:
                txt = FormattedString(txt)
                txt.template = param
            env.globals[opt_name].add_param(key, txt)
        except UndefinedError:
            to_eval.append((opt_name, key, param))

    if len(to_eval) > 0:
        msg = "unable to fully render config\n"
        for item in to_eval:
            msg += "%s:%s '%s'\n" % item
        raise UserWarning(msg)

    # add global filters and test
    env.globals['today'] = lambda: "TODAY"  # TODO

    def is_available(opt_param):
        return opt_param in installed_options(env)

    env.tests['available'] = is_available

    return env


def pkg_env(pkg_cfg):
    """Create a jinja2.Environment from a package configuration

    Args:
        pkg_cfg (dict of str, any): package configuration

    Returns:
        (jinja2.Environment)
    """
    env = create_env(pkg_cfg)

    # load option specific handlers
    for name in pkg_cfg:
        if not name.startswith("_"):
            try:
                opt_handlers = import_module("pkglts.option.%s.handlers" % name)
                if not hasattr(opt_handlers, "environment_extensions"):
                    logger.debug("option %s do not define any extension" % name)
                else:
                    extensions = opt_handlers.environment_extensions(env)
                    for k, v in extensions.items():
                        setattr(env.globals[name], k, v)
            except ImportError:
                raise KeyError("option '%s' does not exists" % name)

    return env


def get_pkg_config(rep="."):
    """Read pkg_cfg file associated to this package.

    Args:
        rep (str): directory to search for info

    Returns:
        (jinja2.Environment): env.globals initialized with pkg_config
    """
    with open(pj(rep, pkglts_dir, pkg_cfg_file), 'r') as f:
        pkg_cfg = json.load(f)

    # update version of pkg_config
    file_version = pkg_cfg['_pkglts'].get('version', 0)
    for i in range(file_version, current_pkg_cfg_version):
        upgrade_pkg_cfg_version(pkg_cfg, i)

    # create jinja2 Environment
    env = pkg_env(pkg_cfg)
    if file_version < current_pkg_cfg_version:
        write_pkg_config(env, rep)

    return env


def write_pkg_config(env, rep="."):
    """Store config associated to this package on disk.

    Args:
        env (jinja2.Environment): current working environment
        rep (str): directory to search for info

    Returns:
        None
    """
    logger.info("write package config")
    pkg_cfg = {}
    for opt_name, cfg in env.globals.items():
        if isinstance(cfg, ConfigSection):
            pkg_cfg[opt_name] = {}
            for key, val in cfg.items():
                if isinstance(val, FormattedString):
                    val = val.template
                pkg_cfg[opt_name][key] = val

    with open(pj(rep, pkglts_dir, pkg_cfg_file), 'w') as f:
        json.dump(pkg_cfg, f, sort_keys=True, indent=2)


def upgrade_pkg_cfg_version(pkg_cfg, version):
    """Upgrade the version of pkg_cfg file from version to version +1

    Args:
        pkg_cfg (dict of str, any): package configuration
        version (int): current version of file

    Returns:
        (dict of str: any): a reference to an updated pkg_cfg
    """
    if version == 0:
        pkg_cfg['_pkglts']['version'] = 1
    elif version == 1:
        pkg_cfg['_pkglts']['version'] = 2
    elif version == 2:
        pkg_cfg['_pkglts']['version'] = 3
        if 'pysetup' in pkg_cfg:
            section = pkg_cfg['pysetup']
            section['require'] = section.get('require', [])

    return pkg_cfg


def installed_options(env):
    """List all installed options according to current environment

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (iter of str)
    """
    for key, cfg in env.globals.items():
        if isinstance(cfg, ConfigSection) and not key.startswith("_"):
            yield key



from importlib import import_module

from pkglts.config_management import installed_options


def requirements(env, requirement_name):
    """Check all requirements for installed options.

    Args:
        env (jinja2.Environment):
        requirement_name (str): type of requirement 'install', 'dvlpt'

    Returns:
        (list of str): list of required packages names
    """
    reqs = set()
    for name in installed_options(env):
        try:
            opt_req = import_module("pkglts.option.%s.require" % name)
            reqs.update(getattr(opt_req, requirement_name))
        except ImportError:
            raise KeyError("option '%s' does not exists" % name)

    return sorted(reqs)


def pkg_url(env):
    try:
        url = env.globals['base'].url
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = env.globals['github'].url
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = env.globals['pypi'].url
        if url is not None:
            return url
    except KeyError:
        pass

    try:
        url = env.globals['readthedocs'].url
        if url is not None:
            return url
    except KeyError:
        pass

    return ""


def environment_extensions(env):
    """Add more functionality to an environment.

    Args:
        env (jinja2.Environment):

    Returns:
        dict of str: any
    """
    req_install = requirements(env, 'install')
    for tup in env.globals['pysetup'].require:
        req_install.append(tup)

    req_dvlpt = requirements(env, 'dvlpt')

    def req(name):
        if name == 'install':
            return req_install
        elif name == 'dvlpt':
            return req_dvlpt
        else:
            raise UserWarning("WTF")

    return {"pkg_url": pkg_url(env),
            "requirements": req}



# rev = 1
""" Contains functions to manage the structure of the package.

Use 'setup.py' for common tasks.
"""

from importlib import import_module
import logging
from os import listdir, mkdir, remove, walk
from os.path import exists, normpath, splitext
from os.path import join as pj
from shutil import rmtree

from .config import pkglts_dir, pkg_cfg_file, pkg_hash_file
from .data_access import get_data_dir, ls
from .config_management import (pkg_env, default_cfg,
                                get_pkg_config, installed_options,
                                write_pkg_config)
from .hash_management import (get_pkg_hash, modified_file_hash,
                              pth_as_key, write_pkg_hash)
from .manage_tools import (check_option_parameters,
                           regenerate_dir, update_opt)
from .option_tools import get_user_permission

logger = logging.getLogger(__name__)


def init_pkg(rep="."):
    """Initialise a package in given directory.

    Args:
        rep (str): directory to create pkg into, default current

    Returns:
        None
    """
    if not exists(pj(rep, pkglts_dir)):
        mkdir(pj(rep, pkglts_dir))

    logger.info("init package")

    for name in ("regenerate.no", "clean.no"):
        if not exists(pj(rep, pkglts_dir, name)):
            with open(pj(rep, pkglts_dir, name), 'w') as f:
                f.write("")

    if exists(pj(rep, pkglts_dir, pkg_cfg_file)):
        env = get_pkg_config(rep)
    else:
        env = pkg_env(default_cfg)
    write_pkg_config(env, rep)

    if not exists(pj(rep, pkglts_dir, pkg_hash_file)):
        write_pkg_hash({}, rep)


def clean(rep="."):
    """Thorough cleaning of all arborescence rooting at rep.

    Todo: exception list instead of hardcoded one

    Args:
        rep (str): default ".", top directory to clean

    Returns:
        None
    """
    for name in ("build", "dist", "doc/_dvlpt", "doc/build"):
        pth = normpath(pj(rep, name))
        if exists(pth):
            rmtree(pth)

    for root, dnames, fnames in walk(rep):
        # do not walk directories starting with "."
        for name in tuple(dnames):
            if "clean.no" in listdir(pj(root, name)):
                dnames.remove(name)
            elif name.startswith("."):
                dnames.remove(name)
            elif name == "__pycache__":
                rmtree(pj(root, name))
                dnames.remove(name)

        for name in fnames:
            if not name.startswith("."):
                if splitext(name)[1] in [".pyc", ".pyo"]:
                    remove(pj(root, name))


def add_option(name, env):
    """Add a new option to this package.

    Notes: See the list of available options online

    Args:
        name (str): name of option to add
        env (jinja2.Environment): current working environment

    Returns:
        (jinja2.Environment): updated environment
    """
    if name in installed_options(env):
        raise UserWarning("option already included in this package")

    return update_opt(name, env)


def install_example_files(option, env, target="."):
    """Install example files associated to an option.

    Args:
        option (str): name of option
        env (jinja2.Environment): current working environment
        target (str): target directory to write into

    Returns:
        None
    """
    if option not in installed_options(env):
        logger.warning("please install option before example files")
        return False

    if (option, True) not in ls("example"):
        logger.info("option does not provide any example")
        return False

    root = pj(get_data_dir(), 'example', option)
    regenerate_dir(root, target, env, {})
    return True


def regenerate_package(env, target=".", overwrite=False):
    """Rebuild all automatically generated files.

    Args:
        env (jinja2.Environment): current working environment
        target (str): target directory to write into
        overwrite (bool): default False, whether or not
                         to overwrite user modified files

    Returns:
        None
    """
    # check consistency of env params
    invalids = []
    for option in installed_options(env):
        for n in check_option_parameters(option, env):
            invalids.append("%s.%s" % (option, n))

    if len(invalids) > 0:
        for param in invalids:
            logger.warning("param %s is not valid", param)

        return False

    # check for potential conflicts
    hm_ref = get_pkg_hash(target)

    conflicted = []
    for file_pth in hm_ref:
        pth = pj(target, file_pth)
        if exists(pth) and modified_file_hash(pth, hm_ref):
            conflicted.append(pth)
        else:
            # file disappeared, regenerate_dir will reload it if managed by pkglts
            pass

    overwrite_file = {}
    if len(conflicted) > 0:
        if overwrite:
            for name in conflicted:
                logger.debug("conflicted, '%s'" % name)
                overwrite_file[pth_as_key(name)] = True
        else:
            for name in conflicted:
                print("A non editable section of %s has been modified" % name)
                overwrite_file[pth_as_key(name)] = get_user_permission("overwrite", False)

    # render files for all options
    hm = {}
    for name in installed_options(env):
        opt_ref_dir = pj(get_data_dir(), 'base', name)
        if not exists(opt_ref_dir):
            logger.debug("option %s do not provide files" % name)
        else:
            logger.info("rendering option %s" % name)
            loc_hm = regenerate_dir(opt_ref_dir, target, env, overwrite_file)
            hm.update(loc_hm)

    hm_ref.update(hm)
    write_pkg_hash(hm_ref, target)


def regenerate_option(env, name, target=".", overwrite=False):
    """Call the regenerate function of a given option

    Args:
        env (jinja2.Environment): current working environment
        name: (str) name of option
        target: (str) target directory to write into
        overwrite (bool): default False, whether or not
                         to overwrite user modified files

    Returns:
        None
    """
    # test existence of option regenerate module
    try:
        import_module("pkglts.option.%s" % name)
    except ImportError:
        raise KeyError("option '%s' does not exists" % name)
    try:
        opt_regenerate = import_module("pkglts.option.%s.regenerate" % name)
    except ImportError:
        raise KeyError("option '%s' does not provide regeneration" % name)

    opt_regenerate.main(env, target, overwrite)



from nose.tools import with_setup
from os import mkdir, rmdir
from os.path import exists

from pkglts.config_management import create_env
from pkglts.option.notebook.config import parameters, check

tmp_dir = "nb_test_config"


def test_parameters():
    assert len(parameters) == 1


def setup_func():
    mkdir(tmp_dir)


def teardown_func():
    if exists(tmp_dir):
        rmdir(tmp_dir)


@with_setup(setup=setup_func, teardown=teardown_func)
def test_config_check_src_directory():
    env = create_env(dict(notebook={'src_directory': "failed_nb"}))
    assert 'src_directory' in check(env)

    env = create_env(dict(notebook={'src_directory': tmp_dir}))
    assert 'src_directory' not in check(env)



from pkglts.config_management import create_env
from pkglts.option.pypi.config import check, parameters


def test_parameters():
    assert len(parameters) == 1


def test_config_check_classifiers_exists():
    env = create_env(dict(pypi={'classifiers': []}))
    assert 'classifiers' in check(env)



import json
import mock
from nose.tools import assert_raises, with_setup
from os import remove
from os.path import exists
from os.path import join as pj

from pkglts.config import pkglts_dir, pkg_cfg_file
from pkglts.config_management import create_env, get_pkg_config, write_pkg_config
from pkglts.manage import (init_pkg, regenerate_package)

from .small_tools import ensure_created, rmdir


tmp_dir = 'toto_mg_rg'
init_file = pj(tmp_dir, "src", "toto", "__init__.py")


def addendum():
    """ modify init_file in first pkglts div
    """
    with open(init_file, 'r') as f:
        lines = f.read().splitlines()

    lines.insert(1, "addendum")

    with open(init_file, 'w') as f:
        f.write("\n".join(lines))


def setup():
    ensure_created(tmp_dir)
    init_pkg(tmp_dir)
    with open(pj(tmp_dir, pkglts_dir, pkg_cfg_file), 'r') as f:
        cfg = json.load(f)

    cfg['base'] = dict(pkgname='toto', namespace=None,
                       authors=[('moi', 'moi@email.com')], url=None)
    env = create_env(cfg)
    write_pkg_config(env, tmp_dir)
    regenerate_package(env, tmp_dir)


def teardown():
    rmdir(tmp_dir)


@with_setup(setup, teardown)
def test_regenerate_pass():
    with open(pj(tmp_dir, pkglts_dir, pkg_cfg_file), 'r') as f:
        cfg = json.load(f)

    cfg['base'] = dict(pkgname='toto', namespace=None,
                       authors=[('moi', 'moi@email.com')], url=None)
    env = create_env(cfg)
    regenerate_package(env, tmp_dir)
    assert exists(init_file)


@with_setup(setup, teardown)
def test_regenerate_check_pkg_cfg_validity():
    with open(pj(tmp_dir, pkglts_dir, pkg_cfg_file), 'r') as f:
        cfg = json.load(f)

    cfg['base'] = dict(pkgname='1toto', namespace=None,
                       authors=[('moi', 'moi@email.com')], url=None)
    env = create_env(cfg)
    assert not regenerate_package(env, tmp_dir)


@with_setup(setup, teardown)
def test_regenerate_handle_conflicts_keep():
    env = get_pkg_config(tmp_dir)

    with open(init_file, 'w') as f:
        f.write("modified")

    with mock.patch('pkglts.manage.get_user_permission',
                    return_value=False):
        regenerate_package(env, tmp_dir)

    with open(init_file, 'r') as f:
        assert f.read() == "modified"


@with_setup(setup, teardown)
def test_regenerate_handle_conflicts_overwrite():
    env = get_pkg_config(tmp_dir)

    addendum()

    with mock.patch('pkglts.manage.get_user_permission',
                    return_value=True):
        regenerate_package(env, tmp_dir)

    with open(init_file, 'r') as f:
        txt = f.read()
        assert "modified" not in txt


@with_setup(setup, teardown)
def test_regenerate_handle_global_overwrite():
    env = get_pkg_config(tmp_dir)

    addendum()

    regenerate_package(env, tmp_dir, overwrite=True)

    with open(init_file, 'r') as f:
        txt = f.read()
        assert "modified" not in txt


@with_setup(setup, teardown)
def test_regenerate_new_files_do_not_generate_conflicts():
    env = get_pkg_config(tmp_dir)

    new_pth = pj(tmp_dir, "src", "toto", "new_file.py")
    with open(new_pth, 'w') as f:
        f.write("txt = 'addendum'")

    regenerate_package(env, tmp_dir)

    assert exists(new_pth)
    with open(new_pth, 'r') as f:
        txt = f.read()
        assert txt == "txt = 'addendum'"


@with_setup(setup, teardown)
def test_regenerate_remove_user_files_do_not_generate_conflicts():
    env = get_pkg_config(tmp_dir)

    new_pth = pj(tmp_dir, "src", "toto", "new_file.py")
    with open(new_pth, 'w') as f:
        f.write("txt = 'addendum'")

    regenerate_package(env, tmp_dir)

    remove(new_pth)

    regenerate_package(env, tmp_dir)
    assert not exists(new_pth)


@with_setup(setup, teardown)
def test_regenerate_fail_if_permanent_section_ids_have_been_modified():
    env = get_pkg_config(tmp_dir)

    with open(init_file, 'a') as f:
        f.write("\n# {# pkglts, test\na = 1\n# #}\n")

    assert_raises(KeyError, lambda: regenerate_package(env, tmp_dir, overwrite=True))



from pkglts.config_management import create_env
from pkglts.option.base.config import check, parameters


def test_parameters():
    assert len(parameters) == 4


def test_config_check_pkg_names():
    for pkg in ('1mypkg', ' mypkg', '1', '1.mypkg',
                ' .mypkg', '.mypkg', 'None.mypkg', 'oa.o.mypkg'):
        env = create_env(dict(base={'pkgname': pkg,
                                    'namespace': None,
                                    'owner': 'moi',
                                    'url': None}))
        assert 'pkgname' in check(env)
        env = create_env(dict(base={'pkgname': 'toto',
                                    'namespace': pkg,
                                    'owner': 'moi',
                                    'url': None}))
        assert 'namespace' in check(env)



# coding: utf-8

from os import listdir
from os.path import isdir
from os.path import join as pj

from pkglts.config_management import create_env


def nn(env, pth):
    tgt_name = env.from_string(pth).render()
    if tgt_name.endswith(".tpl"):
        tgt_name = tgt_name[:-4]

    return tgt_name


def tree(dname, padding, txt):
    pkg_cfg = dict(base={"namespace": None,
                         "owner": "owner",
                         "pkgname": "pkgname",
                         "url": None})

    env = create_env(pkg_cfg)

    files = [(isdir(pj(dname, fname)), fname) for fname in listdir(dname)]
    files.sort()

    count = 0
    for is_dir, fname in files:
        count += 1
        txt += padding + '|\n'
        fmt_name = nn(env, fname)
        txt += padding + '+-' + fmt_name
        path = pj(dname, fname)
        if is_dir:
            txt += "/\n"
            if count == len(files):
                txt = tree(path, padding + ' ' + ' ' * (len(fmt_name) / 2), txt)
            else:
                txt = tree(path, padding + '|' + ' ' * (len(fmt_name) / 2), txt)
        else:
            txt += '\n'

    return txt


def fmt_tree(dname):
    return tree(dname, '', ".\n")



parameters = [
    ("description", "belle petite description"),
    ("keywords", [])
]


def check(env):
    """Check the validity of parameters in working environment.

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    description = env.globals['doc'].description
    # keywords = env.globals['doc'].keywords

    if len(description) == 0:
        invalids.append("description")

    return invalids



import logging
import logging.handlers
# from os import remove, rename
# from os.path import exists


fmt = logging.Formatter('%(asctime)s %(levelname)s (%(name)s): %(message)s')

wng_ch = logging.StreamHandler()
wng_ch.setLevel(logging.WARNING)
wng_ch.setFormatter(fmt)

logger = logging.getLogger("pkglts")
logger.setLevel(logging.DEBUG)
logger.addHandler(wng_ch)

try:
    # n = 5
    # tpl = ".pkglts/info.log.%d"
    # if exists(tpl % n):
    #     remove(tpl % n)
    #
    # for i in range(n, 0, -1):
    #     if exists(tpl % (i - 1)):
    #         rename(tpl % (i - 1), tpl % i)
    #
    # if exists(".pkglts/info.log"):
    #     rename(".pkglts/info.log", tpl % 0)
    #
    # info_ch = logging.FileHandler(".pkglts/info.log", 'w')
    info_ch = logging.StreamHandler()
    info_ch.setLevel(logging.INFO)
    info_ch.setFormatter(fmt)

    logger.addHandler(info_ch)
except IOError:
    print("no info handler")



from pkglts.config_management import create_env
from pkglts.option.license.config import check, parameters


def test_parameters():
    assert len(parameters) == 4


def test_config_check_license_name_exists():
    env = create_env(dict(license={'name': "", 'year': 2015,
                                   'organization': "oa", 'project': "project"}))
    assert 'name' in check(env)

    env = create_env(dict(license={'name': "tugudu", 'year': 2015,
                                   'organization': "oa", 'project': "project"}))
    assert 'name' in check(env)



from pkglts.option.doc import fmt_badge


def environment_extensions(env):
    """Add more functionality to an environment.

    Args:
        env (jinja2.Environment):

    Returns:
        dict of str: any
    """
    owner = env.globals['github'].owner
    project = env.globals['github'].project

    url = "coveralls.io/github/%s/%s?branch=master" % (owner, project)
    img = ("coveralls.io/repos/github/%s/%s/" % (owner, project) +
           "badge.svg?branch=master")
    badge = fmt_badge(img, url, "Coverage report status")

    return {"badge": badge}



from os.path import exists


parameters = [
    ("src_directory", "example")
]


def check(env):
    """Check the validity of parameters in working environment.

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    src_directory = env.globals['notebook'].src_directory

    if not exists(src_directory):
        invalids.append("src_directory")

    return invalids



from argparse import ArgumentParser, RawTextHelpFormatter
import logging

from .config_management import (get_pkg_config, installed_options,
                                write_pkg_config)
from .manage import (clean, init_pkg, install_example_files,
                     regenerate_package, regenerate_option, add_option)

logger = logging.getLogger(__name__)


def action_clean(*args, **kwds):
    """Clean package of all un necessary files.
    """
    del args  # unused
    del kwds  # unused
    logger.info("clean package")
    clean()


def action_init(*args, **kwds):
    """Initialize environment for use of pkglts.
    """
    del args  # unused
    del kwds  # unused
    init_pkg()


def action_clear(*args, **kwds):
    """Attempt to free the package from pkglts interactions.
    """
    del args  # unused
    del kwds  # unused
    logger.info("clear")
    print("TODO")


def action_update(*args, **kwds):
    """Check if a new version of pkglts is available.
    """
    del args  # unused
    del kwds  # unused
    logger.info("update")
    print("TODO")


def action_regenerate(*args, **kwds):
    """Regenerate all files in the package.
    """
    overwrite = 'overwrite' in kwds

    env = get_pkg_config()
    clean()

    if len(args) == 0:
        logger.info("regenerate package")
        regenerate_package(env, overwrite=overwrite)
    else:
        for name in args:
            logger.info("regenerate '%s'" % name)
            regenerate_option(env, name, overwrite=overwrite)


def action_add(*args, **kwds):
    """Add new options in the package.
    """
    del kwds  # unused
    if len(args) == 0:
        raise UserWarning("need to specify at least one option name")

    logger.info("add option")
    env = get_pkg_config()
    for name in args:
        env = add_option(name, env)

    write_pkg_config(env)


def action_remove(*args, **kwds):
    """Remove options from the package.
    """
    del kwds  # unused
    if len(args) == 0:
        raise UserWarning("need to specify at least one option name")

    logger.info("remove option")
    print("TODO")


def action_example(*args, **kwds):
    """Install example files associated with options.
    """
    del kwds  # unused
    if len(args) == 0:
        raise UserWarning("need to specify at least one option name")

    logger.info("install examples")
    env = get_pkg_config()
    for name in args:
        install_example_files(name, env)


action = dict(clean=action_clean,
              init=action_init,
              clear=action_clear,
              update=action_update,
              regenerate=action_regenerate,
              rg=action_regenerate,
              add=action_add,
              remove=action_remove,
              example=action_example)


def main():
    parser = ArgumentParser(description='Package structure manager',
                            formatter_class=RawTextHelpFormatter)

    act_help = "type of action performed by pmg, one of:\n"
    for name, func in action.items():
        act_help += "\n  - %s: %s" % (name, func.__doc__)

    parser.add_argument('action', metavar='action',
                        choices=tuple(action.keys()),
                        help=act_help)

    parser.add_argument('action_args', nargs='*',
                        help="action to perform on the package")

    parser.add_argument('-e', metavar='extra', nargs=2, action='append',
                        help='extra arguments to pass to the action',
                        dest='extra')

    args = parser.parse_args()
    if args.extra is None:
        extra = {}
    else:
        extra = dict(args.extra)

    action[args.action](*args.action_args, **extra)


if __name__ == '__main__':
    main()



""" Some helpers for options
"""
try:
    loc_input = raw_input
except NameError:
    loc_input = input


def get_user_permission(action_name, default_true=True):
    if default_true:
        return loc_input("%s [y], n?" % action_name) in ("", "y")
    else:
        return loc_input("%s y, [n]?" % action_name) == "y"

#
# def get_key(key, env):
#     """ Fetch a specific key in env
#     """
#     try:
#         elms = key.split(".")
#         d = env
#         for k in elms:
#             d = d[k]
#
#         return d
#     except KeyError:
#         return None
#
#
# def _ask_single_arg(name, default):
#     # prompt user for value
#     msg = name
#     if default is None:
#         msg += ":"
#         val = ""
#     else:
#         val = str(default)
#         msg += " [%s]:" % val
#
#     ans = loc_input(msg)
#     if ans == "":
#         return val
#     else:
#         return ans
#
#
# def ask_arg(name, pkg_cfg=None, default=None, extra=None):
#     """ Prompt the user for the value of some argument
#
#     If user returns nothing then default will be returned
#
#     If pkg_cfg is provided, the function will attempt to
#     find a proper default in it.
#
#     If extra is provided, the function will attempt to
#     find a value in it and return it without prompting
#     the user.
#
#     If default is a list of elements, even empty, the function
#     will return a list of answers.
#
#     args:
#      - name (str): name of argument. Can be in the form of
#                    'option.arg'.
#      - pkg_cfg (dict of (str, any)): place to look for defaults
#      - default (any): default to use as a last resort
#      - extra (dict of (str, any)): place to look for values
#
#     return:
#      - (str): value of argument either from prompt or from default
#     """
#     if extra is None:
#         extra = {}
#
#     # try to find value in extra
#     if name in extra:
#         return extra[name]
#
#     key = name.split(".")[-1]
#     if key in extra:
#         return extra[key]
#
#     if pkg_cfg is None:
#         pkg_cfg = {}
#
#     if isinstance(default, (list, tuple)):
#         # ask multiple args
#         val = get_key(name, pkg_cfg)
#         if val is None:
#             val = default
#
#         if not isinstance(val, (list, tuple)):
#             msg = "bad value in pkg_cfg, expected list for %s" % name
#             raise UserWarning(msg)
#
#         i = 0
#         items = []
#         res = "txt"
#         while res != "":
#             if i < len(val):
#                 loc_val = val[i]
#             else:
#                 loc_val = ""
#             res = _ask_single_arg(name, loc_val)
#             if res != "":
#                 items.append(res)
#                 i += 1
#
#         return items
#     else:
#         val = get_key(name, pkg_cfg)
#         if val is None:
#             val = default
#         return _ask_single_arg(name, val)



from nose.tools import assert_raises

from pkglts.config_management import create_env, default_cfg
from pkglts.manage import add_option


def test_manage_add_opt_raise_error_if_already_installed():
    env = create_env(default_cfg)
    env = add_option("base", env)
    assert_raises(UserWarning, lambda: add_option('base', env))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# {# pkglts, pysetup.kwds
# format setup arguments

from os import walk
from os.path import abspath, normpath
from os.path import join as pj

from setuptools import setup, find_packages


short_descr = "Building packages with long term support"
readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')


# find version number in src/pkglts/version.py
version = {}
with open("src/pkglts/version.py") as fp:
    exec(fp.read(), version)


data_files = []

nb = len(normpath(abspath("src/pkglts_data"))) + 1


def data_rel_pth(pth):
    """ Return path relative to pkg_data
    """
    abs_pth = normpath(abspath(pth))
    return abs_pth[nb:]


for root, dnames, fnames in walk("src/pkglts_data"):
    for name in fnames:
        data_files.append(data_rel_pth(pj(root, name)))


setup_kwds = dict(
    name='pkglts',
    version=version["__version__"],
    description=short_descr,
    long_description=readme + '\n\n' + history,
    author="revesansparole, ",
    author_email="revesansparole@gmail.com, ",
    url='https://github.com/revesansparole/pkglts',
    license='CeCILL-C',
    zip_safe=False,

    packages=find_packages('src'),
    package_dir={'': 'src'},
    
    include_package_data=True,
    package_data={'pkglts_data': data_files},
    install_requires=[
        ],
    tests_require=[
        "coverage",
        "flake8",
        "mock",
        "nbconvert",
        "nose",
        "sphinx",
        "coveralls",
        "tox",
        "twine",
        ],
    entry_points={},
    keywords='packaging, package builder',
    
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
    ],
    test_suite='nose.collector',
)
# #}
# change setup_kwds below before the next pkglts tag

setup_kwds['entry_points']['console_scripts'] = ['pmg = pkglts.manage_script:main']

# do not change things below
# {# pkglts, pysetup.call
setup(**setup_kwds)
# #}



parameters = []


def check(env):
    """Check the validity of parameters in working environment.

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (list of str): list of faulty parameters
    """
    del env
    return []



from base64 import b64encode
from hashlib import sha512
import json
import logging
from os.path import join as pj
from os.path import normpath

from .config import pkglts_dir, pkg_hash_file
from .templating import parse_source


logger = logging.getLogger(__name__)


def pth_as_key(pth):
    """Normalize path to enable to use them as keys

    Args:
        pth (str): path to normalize

    Returns:
        (str)
    """
    return normpath(pth).replace("\\", "/")


def compute_hash(txt):
    """Compute hash summary of a text

    Args:
        txt (str): content to hash

    Returns:
        (str): hash key
    """
    algo = sha512()
    algo.update(txt.encode('utf-8'))  # TODO bad if non utf-8 encoded file
    return b64encode(algo.digest()).decode('utf-8')


def get_pkg_hash(rep="."):
    """Read pkg_hash file associated to this package

    Args:
        rep (str): directory to search for info

    Returns:
        (dict of str, list): hash map of preserved sections in this
                             package
    """
    with open(pj(rep, pkglts_dir, pkg_hash_file), 'r') as f:
        hm = json.load(f)

    return hm


def write_pkg_hash(pkg_hash, rep="."):
    """Store hash associated to this package on disk.

    Args:
        pkg_hash (dict of str, list): hash map of preserved sections in this
                                     package
        rep (str): directory to search for info

    Returns:
        None
    """
    logger.info("write package hash")
    cfg = dict(pkg_hash)

    with open(pj(rep, pkglts_dir, pkg_hash_file), 'w') as f:
        json.dump(cfg, f, sort_keys=True, indent=2)


def modified_file_hash(pth, pkg_hash):
    """Check whether a file complies with previously stored hash

    Args:
        pth (str): path to file to test
        pkg_hash (dict of str, list): hash map of preserved sections in this
                                     package

    Returns:
        (bool): whether this file has been modified or not
    """
    key = pth_as_key(pth)
    if key not in pkg_hash:
        raise IOError("%s not under check" % key)

    ref_blocks = pkg_hash[key]

    with open(pth, 'r') as f:
        blocks = parse_source(f.read())

    lts_blocks = dict((bid, cnt) for bid, bef, cnt, aft in blocks
                      if bid is not None)
    if set(lts_blocks) != set(ref_blocks):
        return True
    else:
        for bid, cnt in lts_blocks.items():
            sha = compute_hash(cnt)
            if sha != ref_blocks[bid]:
                return True




parameters = [
    ("theme", 'default'),
    ("autodoc_dvlpt", True)
]


def check(env):
    """Check the validity of parameters in working environment.

    Args:
        env (jinja2.Environment):  current working environment

    Returns:
        (list of str): list of faulty parameters
    """
    invalids = []
    theme = env.globals['sphinx'].theme
    if theme != str(theme):
        invalids.append('theme')

    return invalids



from os.path import exists
import re

opening_marker = "{" + "#"
closing_marker = "#" + "}"

block_re = re.compile(r"\{#[ ]pkglts,[ ](?P<key>.*?)\n(?P<cnt>.*?)#\}",
                      re.DOTALL | re.MULTILINE)


def parse_source(txt):
    """Parse text to find preserved blocks

    Args:
        txt (str): full text to parse

    Returns:
        (list of [str, str, str, str]): ordered list of blocks:
                    - block_id (or None if content is not preserved
                    - line start before start for preserved content
                    - content
                    - line start before end for preserved content
    """
    blocks = []
    last_end = 0

    for res in block_re.finditer(txt):
        i = res.start()
        while i > last_end and txt[i] != "\n":
            i -= 1

        if i >= last_end:
            blocks.append((None, "", txt[last_end: (i + 1)], ""))

        bef = txt[(i + 1): res.start()]

        bid = res.group('key')

        cnt = res.group('cnt')

        i = len(cnt) - 1
        while i > 0 and cnt[i] != "\n":
            i -= 1
        aft = cnt[(i + 1): len(cnt)]
        cnt = cnt[:i]
        last_end = res.end() + 1

        blocks.append((bid, bef, cnt, aft))

    if last_end < len(txt):
        blocks.append((None, "", txt[last_end: len(txt)], ""))

    return blocks


def render(env, src_pth, tgt_pth):
    """Render src_pth templated file into tgt_pth

    Notes: keeps 'preserved' block structure

    Args:
        env (jinja2.Environment): current pkg environment
        src_pth (str): path to reference file
        tgt_pth (str): path to potentially non existent yet target file

    Returns:
        (list of [str, str]): key, cnt for preserved blocks
    """
    # parse src file to find 'preserved' blocks
    with open(src_pth, 'r') as f:
        src_blocks = parse_source(f.read())

    blocks = []
    if exists(tgt_pth):  # retrieves preserved blocks from source
        # parse tgt file to find 'preserved' blocks
        with open(tgt_pth, 'r') as f:
            tgt_blocks = parse_source(f.read())

        src_blocks = dict((bid, cnt) for (bid, bef, cnt, aft) in src_blocks
                          if bid is not None)
        for bid, bef, cnt, aft in tgt_blocks:
            # check stored content hash

            if bid is not None:
                cnt = src_blocks[bid]

            blocks.append((bid, bef, cnt, aft))
    else:  # format non preserved blocks for the first and only time
        for bid, bef, cnt, aft in src_blocks:
            if bid is None:
                template = env.from_string(cnt)
                cnt = template.render()

            blocks.append((bid, bef, cnt, aft))

    # regenerate preserved block content
    # print "blocks", [bid for bid, bef, cnt, aft in blocks]
    preserved = []
    tgt = ""
    for bid, bef, cnt, aft in blocks:
        if bid is None:
            tgt += cnt
        else:
            # format cnt
            template = env.from_string(cnt)
            cnt = template.render()
            preserved.append((bid, cnt))
            # rewrite preserved tag if necessary
            tgt += bef + "{" + "# pkglts, %s\n" % bid
            tgt += cnt
            tgt += "\n" + aft + "#" + "}\n"

    with open(tgt_pth, 'w') as f:
        f.write(tgt)

    return preserved



from pkglts.config_management import create_env
from pkglts.option.doc.config import check, parameters


def test_parameters():
    assert len(parameters) == 2


def test_config_check_description_exists():
    env = create_env(dict(doc={'description': "mydescr", 'keywords': []}))
    assert env.globals['doc'].description == "mydescr"
    assert 'description' not in check(env)


def test_config_check_description_valid():
    env = create_env(dict(doc={'description': "", 'keywords': []}))
    assert 'description' in check(env)


def test_config_check_keywords_exists():
    env = create_env(dict(doc={'description': "mydescr", 'keywords': []}))
    assert len(env.globals['doc'].keywords) == 0
    assert 'keywords' not in check(env)
